from unittest import mock

from .mocked_producer import MockedProducer


class RelayerPatch(object):

    def __init__(self):
        self.patcher = mock.patch('relayer.KafkaProducer')

    def start(self):
        kafka_producer_mock = self.patcher.start()
        self.mocked_producer = MockedProducer()
        kafka_producer_mock.return_value = self.mocked_producer

    def stop(self):
        self.patcher.stop()



from relayer import Relayer
from .flask_context_handler import FlaskContextHandler
from .logging_middleware import LoggingMiddleware


class FlaskRelayer(object):

    def __init__(self, app=None, logging_topic=None, kafka_hosts=None, topic_prefix='', topic_suffix='', source=''):
        if app:
            self.init_app(app, logging_topic, kafka_hosts=kafka_hosts, topic_prefix=topic_prefix,
                          topic_suffix=topic_suffix, source=source)

    def init_app(self, app, logging_topic, kafka_hosts=None, topic_prefix='', topic_suffix='', source=''):
        kafka_hosts = kafka_hosts or app.config.get('KAFKA_HOSTS')
        self.event_relayer = Relayer(logging_topic, FlaskContextHandler, kafka_hosts=kafka_hosts, topic_prefix=topic_prefix,
                                     topic_suffix=topic_suffix, source=source)
        app.wsgi_app = LoggingMiddleware(app, app.wsgi_app, self.event_relayer, logging_topic)

    def emit(self, *args, **kwargs):
        self.event_relayer.emit(*args, **kwargs)

    def emit_raw(self, *args, **kwargs):
        self.event_relayer.emit_raw(*args, **kwargs)

    def log(self, *args, **kwargs):
        self.event_relayer.log(*args, **kwargs)



import logging


logger = logging.getLogger('relayer')


def log_kafka_message(topic, payload, partition_key=None):
    logger.debug('topic: %s, partition_key: %s, message: %s', topic, partition_key, payload)



from relayer import Relayer, utils
from .rpc_context_handler import RPCContextHandler

from datetime import datetime


def make_rpc_relayer(logging_topic, kafka_hosts=None, topic_prefix='', topic_suffix='', source=''):

    event_relayer = Relayer(logging_topic, RPCContextHandler, kafka_hosts=kafka_hosts,
                            topic_prefix=topic_prefix, topic_suffix=topic_suffix, source=source)
    context = event_relayer.context

    def decorator(function):
        def wrapper(*args, **kwargs):
            start_time = datetime.utcnow()
            kwargs['relayer'] = event_relayer
            context.start_request()
            service_response = function(*args, **kwargs)
            end_time = datetime.utcnow()
            request_log = {
                'source': event_relayer.source,
                'logging_topic': logging_topic,
                'date': start_time.isoformat(),
                'service': function.__qualname__,
                'service_time': utils.get_elapsed_time_in_milliseconds(start_time, end_time)
            }
            context.end_request(logging_topic, request_log)
            return service_response
        return wrapper

    # Expose relayer instance
    decorator.instance = event_relayer

    return decorator



from kafka import KafkaProducer

from .event_emitter import EventEmitter
from .exceptions import ConfigurationError

__version__ = '0.1.1'


class Relayer(object):

    def __init__(self, logging_topic, context_handler_class, kafka_hosts=None, topic_prefix='', topic_suffix='', source=''):
        self.logging_topic = logging_topic
        if not kafka_hosts:
            raise ConfigurationError()
        if source == '':
            self.source = '{0}{1}{2}'.format(topic_prefix, logging_topic, topic_suffix)
        else:
            self.source = source
        producer = KafkaProducer(bootstrap_servers=kafka_hosts)
        emitter = EventEmitter(producer, topic_prefix=topic_prefix, topic_suffix=topic_suffix)
        self.context = context_handler_class(emitter)

    def emit(self, event_type, event_subtype, payload, partition_key=None):
        payload = {
            'source': self.source,
            'event_type': event_type,
            'event_subtype': event_subtype,
            'payload': payload
        }
        self.context.emit(event_type, payload, partition_key)

    def emit_raw(self, topic, message, partition_key=None):
        self.context.emit(topic, message, partition_key)

    def log(self, log_level, payload):
        message = {
            'log_level': log_level,
            'payload': payload
        }
        self.context.log(message)



class RPCContextHandler(object):
    def __init__(self, event_emitter):
        self.event_emitter = event_emitter
        self.kafka_producer_request_logs = []

    def start_request(self):
        self.kafka_producer_request_logs = []

    def emit(self, topic, message, partition_key=None):
        self.event_emitter.emit(topic, message, partition_key)
        # It is important to first emit the message before appending to the log list so if the emit call fails because
        # the messsage is malformed, the final log message doesn't fail for the same reason.
        self.kafka_producer_request_logs.append(message)

    def log(self, message):
        self.kafka_producer_request_logs.append(message)

    def end_request(self, logging_topic, request_log):
        log_entry = {}
        log_entry.update(request_log)
        log_entry['lines'] = self.kafka_producer_request_logs

        self.event_emitter.emit(logging_topic, log_entry)
        self.event_emitter.flush()



import json
from uuid import UUID

from .exceptions import NonJSONSerializableMessageError, UnsupportedPartitionKeyTypeError
from .logger import log_kafka_message


class EventEmitter(object):

    def __init__(self, producer, topic_prefix='', topic_suffix=''):
        self.producer = producer
        self.topic_prefix = topic_prefix
        self.topic_suffix = topic_suffix

    def emit(self, topic, message, partition_key=None):

        topic = '{0}{1}{2}'.format(self.topic_prefix, topic, self.topic_suffix)

        log_kafka_message(topic, message, partition_key=partition_key)

        if isinstance(partition_key, str):
            partition_key = partition_key.encode('utf-8')
        elif isinstance(partition_key, UUID):
            partition_key = partition_key.bytes
        elif partition_key is not None:
            raise UnsupportedPartitionKeyTypeError(partition_key.__class__)

        try:
            message = json.dumps(message).encode('utf-8')
        except TypeError as error:
            raise NonJSONSerializableMessageError(str(error))

        self.producer.send(topic, key=partition_key, value=message)

    def flush(self):
        self.producer.flush()



import re
from pip.req import parse_requirements
from setuptools import setup, find_packages


def requirements(filename):
    reqs = parse_requirements(filename, session=False)
    return [str(r.req) for r in reqs]


def get_version():
    with open('relayer/__init__.py', 'r') as f:
        version_regex = r'^__version__\s*=\s*[\'"](.+)[\'"]'
        return re.search(version_regex, f.read(), re.MULTILINE).group(1)

setup(
    name='relayer',
    version=get_version(),
    url='https://github.com/wizeline/relayer',
    author='Wizeline',
    author_email='engineering@wizeline.com',
    description='Relayer is a library to emit kafka messages and group logs to relay them to kafka for log aggregation.',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    keywords=['relayer', 'log', 'kafka'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3'
    ],
    tests_require=requirements('requirements-dev.txt'),
    install_requires=requirements('requirements.txt'),
    extras_require={
        'dev': requirements('requirements-dev.txt')
    }
)



from uuid import UUID


class RelayerError(Exception):
    pass


class NonJSONSerializableMessageError(RelayerError):
    pass


class ConfigurationError(RelayerError):
    def __init__(self):
        super().__init__('Cannot create relayer object if no kafka hosts are provided')


class UnsupportedPartitionKeyTypeError(RelayerError):
    def __init__(self, provided_type):
        super().__init__('Unsuported partition key type provided, got {}, expected {} or {}'.format(
            provided_type.__name__, str.__name__, UUID.__name__))



from datetime import datetime


class LoggingMiddleware(object):
    def __init__(self, app, wsgi_app, relayer, logging_topic):
        self.app = app
        self.wsgi_app = wsgi_app
        self.relayer = relayer
        self.logging_topic = logging_topic

    def __call__(self, environ, start_response):
        with self.app.app_context():
            start_time = datetime.utcnow()
            status_code = None
            content_length = None

            self.relayer.context.start_request()

            def logging_start_response(status, response_headers, exc_info=None):
                nonlocal status_code, content_length
                status_code = int(status.partition(' ')[0])
                for name, value in response_headers:
                    if name.lower() == 'content-length':
                        content_length = int(value)
                        break
                return start_response(status, response_headers, exc_info)

            response = self.wsgi_app(environ, logging_start_response)

            if content_length is None:
                content_length = len(b''.join(response))

            elapsed_time = datetime.utcnow() - start_time
            elapsed_time_milliseconds = elapsed_time.microseconds / 1000.0 + elapsed_time.seconds * 1000

            request_log = {
                'source': self.relayer.source,
                'logging_topic': self.logging_topic,
                'date': start_time.isoformat(),
                'user_agent': environ.get('HTTP_USER_AGENT'),
                'method': environ.get('REQUEST_METHOD'),
                'path': environ.get('PATH_INFO'),
                'query_string': environ.get('QUERY_STRING'),
                'remote_addr': environ.get('HTTP_X_REAL_IP', environ.get('REMOTE_ADDR')),
                'x_forwarded_for': environ.get('HTTP_X_FORWARDED_FOR'),
                'status': status_code,
                'content_length': content_length,
                'request_time': elapsed_time_milliseconds
            }

            self.relayer.context.end_request(self.logging_topic, request_log)

            return response



from collections import defaultdict


class MockedProducer(object):
    def __init__(self):
        self.produced_messages = defaultdict(list)
        self.flushed = False

    def send(self, topic, value=None, key=None):
        self.produced_messages[topic].append((value, key, ))

    def flush(self):
        self.flushed = True
